import click

from dbt_platform_helper.domain.codebase import Codebase
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.utils.click import ClickDocOptGroup
from dbt_platform_helper.utils.versioning import (
    check_platform_helper_version_needs_update,
)


@click.group(chain=True, cls=ClickDocOptGroup)
def codebase():
    """Codebase commands."""
    check_platform_helper_version_needs_update()


@codebase.command()
def prepare():
    """Sets up an application codebase for use within a DBT platform project."""
    try:
        Codebase().prepare()
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))


@codebase.command()
@click.option("--app", help="AWS application name", required=True)
@click.option(
    "--with-images",
    help="List up to the last 10 images tagged for this codebase",
    default=False,
    is_flag=True,
)
def list(app, with_images):
    """List available codebases for the application."""
    try:
        Codebase().list(app, with_images)
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))


@codebase.command()
@click.option("--app", help="AWS application name", required=True)
@click.option(
    "--codebase",
    help="The codebase name as specified in the platform-config.yml file",
    required=True,
)
@click.option("--commit", help="GitHub commit hash", required=True)
def build(app, codebase, commit):
    """Trigger a CodePipeline pipeline based build."""
    try:
        Codebase().build(app, codebase, commit)
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))


@codebase.command()
@click.option("--app", help="AWS application name", required=True)
@click.option("--env", help="AWS Copilot environment", required=True)
@click.option(
    "--codebase",
    help="The codebase name as specified in the platform-config.yml file",
    required=True,
)
@click.option("--commit", help="GitHub commit hash", required=True)
def deploy(app, env, codebase, commit):
    try:
        Codebase().deploy(app, env, codebase, commit)
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))



from collections.abc import Callable
from os import makedirs
from pathlib import Path
from shutil import rmtree

from dbt_platform_helper.constants import CODEBASE_PIPELINES_KEY
from dbt_platform_helper.constants import ENVIRONMENT_PIPELINES_KEY
from dbt_platform_helper.constants import ENVIRONMENTS_KEY
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.files import FileProvider
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.utils.application import get_application_name
from dbt_platform_helper.utils.aws import get_account_details
from dbt_platform_helper.utils.aws import get_public_repository_arn
from dbt_platform_helper.utils.files import generate_override_files_from_template
from dbt_platform_helper.utils.template import setup_templates
from dbt_platform_helper.utils.versioning import (
    get_required_terraform_platform_modules_version,
)


class Pipelines:
    def __init__(
        self,
        config_provider: ConfigProvider,
        get_git_remote: Callable[[], str],
        get_codestar_arn: Callable[[str], str],
        io: ClickIOProvider = ClickIOProvider(),
    ):
        self.config_provider = config_provider
        self.get_git_remote = get_git_remote
        self.get_codestar_arn = get_codestar_arn
        self.io = io

    def generate(self, terraform_platform_modules_version, deploy_branch):
        pipeline_config = self.config_provider.load_and_validate_platform_config()

        has_codebase_pipelines = CODEBASE_PIPELINES_KEY in pipeline_config
        has_environment_pipelines = ENVIRONMENT_PIPELINES_KEY in pipeline_config

        if not (has_codebase_pipelines or has_environment_pipelines):
            self.io.warn("No pipelines defined: nothing to do.")
            return

        platform_config_terraform_modules_default_version = pipeline_config.get(
            "default_versions", {}
        ).get("terraform-platform-modules", "")

        templates = setup_templates()
        app_name = get_application_name()

        git_repo = self.get_git_remote()
        if not git_repo:
            self.io.abort_with_error("The current directory is not a git repository")

        codestar_connection_arn = self.get_codestar_arn(app_name)
        if codestar_connection_arn is None:
            self.io.abort_with_error(f'There is no CodeStar Connection named "{app_name}" to use')

        base_path = Path(".")
        copilot_pipelines_dir = base_path / f"copilot/pipelines"

        self._clean_pipeline_config(copilot_pipelines_dir)

        if has_environment_pipelines:
            environment_pipelines = pipeline_config[ENVIRONMENT_PIPELINES_KEY]

            for config in environment_pipelines.values():
                aws_account = config.get("account")
                self._generate_terraform_environment_pipeline_manifest(
                    pipeline_config["application"],
                    aws_account,
                    terraform_platform_modules_version,
                    platform_config_terraform_modules_default_version,
                    deploy_branch,
                )

        if has_codebase_pipelines:
            account_id, _ = get_account_details()

            for codebase in pipeline_config[CODEBASE_PIPELINES_KEY]:
                self._generate_codebase_pipeline(
                    account_id,
                    app_name,
                    codestar_connection_arn,
                    git_repo,
                    codebase,
                    base_path,
                    copilot_pipelines_dir,
                    templates,
                )

    def _clean_pipeline_config(self, pipelines_dir):
        if pipelines_dir.exists():
            self.io.info("Deleting copilot/pipelines directory.")
            rmtree(pipelines_dir)

    def _generate_codebase_pipeline(
        self,
        account_id,
        app_name,
        codestar_connection_arn,
        git_repo,
        codebase,
        base_path,
        pipelines_dir,
        templates,
    ):
        makedirs(pipelines_dir / codebase["name"] / "overrides", exist_ok=True)
        environments = []
        for pipelines in codebase["pipelines"]:
            environments += pipelines[ENVIRONMENTS_KEY]

        additional_ecr = codebase.get("additional_ecr_repository", None)
        add_public_perms = additional_ecr and additional_ecr.startswith("public.ecr.aws")
        additional_ecr_arn = get_public_repository_arn(additional_ecr) if add_public_perms else None

        template_data = {
            "account_id": account_id,
            "app_name": app_name,
            "deploy_repo": git_repo,
            "codebase": codebase,
            ENVIRONMENTS_KEY: environments,
            "codestar_connection_arn": codestar_connection_arn,
            "codestar_connection_id": codestar_connection_arn.split("/")[-1],
            "additional_ecr_arn": additional_ecr_arn,
        }

        self._create_file_from_template(
            base_path,
            f"{codebase['name']}/manifest.yml",
            pipelines_dir,
            template_data,
            templates,
            "codebase/manifest.yml",
        )

        overrides_path = Path(__file__).parent.parent.joinpath(
            "templates/pipelines/codebase/overrides"
        )
        generate_override_files_from_template(
            base_path, overrides_path, pipelines_dir / codebase["name"] / "overrides", template_data
        )

    def _create_file_from_template(
        self, base_path, file_name, pipelines_dir, template_data, templates, template_name=None
    ):
        contents = templates.get_template(
            f"pipelines/{file_name if template_name is None else template_name}"
        ).render(template_data)
        message = FileProvider.mkfile(
            base_path, pipelines_dir / file_name, contents, overwrite=True
        )
        self.io.info(message)

    def _generate_terraform_environment_pipeline_manifest(
        self,
        application,
        aws_account,
        cli_terraform_platform_modules_version,
        platform_config_terraform_modules_default_version,
        deploy_branch,
    ):
        env_pipeline_template = setup_templates().get_template("environment-pipelines/main.tf")

        terraform_platform_modules_version = get_required_terraform_platform_modules_version(
            cli_terraform_platform_modules_version,
            platform_config_terraform_modules_default_version,
        )

        contents = env_pipeline_template.render(
            {
                "application": application,
                "aws_account": aws_account,
                "terraform_platform_modules_version": terraform_platform_modules_version,
                "deploy_branch": deploy_branch,
            }
        )

        dir_path = f"terraform/environment-pipelines/{aws_account}"
        makedirs(dir_path, exist_ok=True)

        self.io.info(FileProvider.mkfile(".", f"{dir_path}/main.tf", contents, overwrite=True))

    def generate_terraform_codebase_pipeline_manifest(
        self,
        application,
        aws_account,
        cli_terraform_platform_modules_version,
        platform_config_terraform_modules_default_version,
        deploy_branch,
    ):
        env_pipeline_template = setup_templates().get_template("codebase-pipelines/main.tf")

        terraform_platform_modules_version = get_required_terraform_platform_modules_version(
            cli_terraform_platform_modules_version,
            platform_config_terraform_modules_default_version,
        )

        contents = env_pipeline_template.render(
            {
                "application": application,
                "aws_account": aws_account,
                "terraform_platform_modules_version": terraform_platform_modules_version,
                "deploy_branch": deploy_branch,
            }
        )

        dir_path = f"terraform/environment-pipelines/{aws_account}"
        makedirs(dir_path, exist_ok=True)

        self.io.info(FileProvider.mkfile(".", f"{dir_path}/main.tf", contents, overwrite=True))



from unittest.mock import Mock

from dbt_platform_helper.constants import DEFAULT_TERRAFORM_PLATFORM_MODULES_VERSION
from dbt_platform_helper.domain.terraform_environment import (
    PlatformTerraformManifestGenerator,
)
from dbt_platform_helper.providers.files import FileProvider


class TestPlatformTerraformManifestGenerator:

    def test_generator_generates_expected_manifest_content_with_version_override(self):
        test_environment_config = {
            "vpc": "vpc3",
            "accounts": {
                "deploy": {"name": "non-prod-acc", "id": "1122334455"},
                "dns": {"name": "non-prod-dns-acc", "id": "6677889900"},
            },
            "versions": {"terraform-platform-modules": 3},
        }
        expected_header = "# WARNING: This is an autogenerated file, not for manual editing."
        expected_modules = "git::https://github.com/uktrade/terraform-platform-modules.git//extensions?depth=1&ref=123456"
        expected_moved_block = (
            "moved {\n  from = module.extensions-tf\n  to   = module.extensions\n}\n"
        )

        result = PlatformTerraformManifestGenerator(Mock()).generate_manifest(
            "test", "test-app", test_environment_config, 123456
        )

        assert expected_header in result
        assert expected_modules in result
        assert expected_moved_block in result
        assert 'environment = "test"' in result
        assert 'application    = "test-app"' in result
        assert 'vpc_name    = "vpc3"'

    def test_generator_generates_expected_manifest_content_with_tpm_version_set_in_config(self):
        test_environment_config = {
            "vpc": "vpc3",
            "accounts": {
                "deploy": {"name": "non-prod-acc", "id": "1122334455"},
                "dns": {"name": "non-prod-dns-acc", "id": "6677889900"},
            },
            "versions": {"terraform-platform-modules": 3},
        }
        expected_header = "# WARNING: This is an autogenerated file, not for manual editing."
        expected_modules = "git::https://github.com/uktrade/terraform-platform-modules.git//extensions?depth=1&ref=3"
        expected_moved_block = (
            "moved {\n  from = module.extensions-tf\n  to   = module.extensions\n}\n"
        )

        result = PlatformTerraformManifestGenerator(Mock()).generate_manifest(
            "test", "test-app", test_environment_config
        )

        assert expected_header in result
        assert expected_modules in result
        assert expected_moved_block in result
        assert 'environment = "test"' in result
        assert 'application    = "test-app"' in result
        assert 'vpc_name    = "vpc3"'

    def test_generator_generates_expected_manifest_content_with_default_version(self):
        test_environment_config = {
            "vpc": "vpc3",
            "accounts": {
                "deploy": {"name": "non-prod-acc", "id": "1122334455"},
                "dns": {"name": "non-prod-dns-acc", "id": "6677889900"},
            },
        }
        expected_header = "# WARNING: This is an autogenerated file, not for manual editing."
        expected_modules = f"git::https://github.com/uktrade/terraform-platform-modules.git//extensions?depth=1&ref={DEFAULT_TERRAFORM_PLATFORM_MODULES_VERSION}"
        expected_moved_block = (
            "moved {\n  from = module.extensions-tf\n  to   = module.extensions\n}\n"
        )

        result = PlatformTerraformManifestGenerator(Mock()).generate_manifest(
            "test", "test-app", test_environment_config
        )

        assert expected_header in result
        assert expected_modules in result
        assert expected_moved_block in result
        assert 'environment = "test"' in result
        assert 'application    = "test-app"' in result
        assert 'vpc_name    = "vpc3"'

    def test_generator_write_manifest_makes_the_expected_manifest_file(self):
        mock_file_provider = Mock(spec=FileProvider)
        PlatformTerraformManifestGenerator(mock_file_provider).write_manifest(
            "test-environment", "test-manifest-content"
        )

        mock_file_provider.mkfile.assert_called_once_with(
            ".",
            f"terraform/environments/test-environment/main.tf",
            "test-manifest-content",
            overwrite=True,
        )



import click

from dbt_platform_helper.constants import DEFAULT_TERRAFORM_PLATFORM_MODULES_VERSION
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.files import FileProvider
from dbt_platform_helper.utils.template import setup_templates


class PlatformTerraformManifestGenerator:
    def __init__(self, file_provider):
        self.file_provider = file_provider
        self.manifest_template = setup_templates().get_template("environments/main.tf")

    def generate_manifest(
        self,
        environment_name: str,
        application_name: str,
        environment_config: dict,
        terraform_platform_modules_version_override: str = None,
    ):
        terraform_platform_modules_version = (
            terraform_platform_modules_version_override
            or environment_config.get("versions", {}).get(
                "terraform-platform-modules", DEFAULT_TERRAFORM_PLATFORM_MODULES_VERSION
            )
        )

        return self.manifest_template.render(
            {
                "application": application_name,
                "environment": environment_name,
                "config": environment_config,
                "terraform_platform_modules_version": terraform_platform_modules_version,
            }
        )

    def write_manifest(self, environment_name: str, manifest_content: str):
        return self.file_provider.mkfile(
            ".",
            f"terraform/environments/{environment_name}/main.tf",
            manifest_content,
            overwrite=True,
        )


class TerraformEnvironmentException(PlatformException):
    pass


class EnvironmentNotFoundException(TerraformEnvironmentException):
    pass


class TerraformEnvironment:
    def __init__(
        self,
        config_provider,
        manifest_generator: PlatformTerraformManifestGenerator = None,
        echo=click.echo,
    ):
        self.echo = echo
        self.config_provider = config_provider
        self.manifest_generator = manifest_generator or PlatformTerraformManifestGenerator(
            FileProvider()
        )

    def generate(self, environment_name, terraform_platform_modules_version_override=None):
        config = self.config_provider.get_enriched_config()

        if environment_name not in config.get("environments").keys():
            raise EnvironmentNotFoundException(
                f"cannot generate terraform for environment {environment_name}.  It does not exist in your configuration"
            )

        manifest = self.manifest_generator.generate_manifest(
            environment_name=environment_name,
            application_name=config["application"],
            environment_config=config["environments"][environment_name],
            terraform_platform_modules_version_override=terraform_platform_modules_version_override,
        )

        self.echo(
            self.manifest_generator.write_manifest(
                environment_name=environment_name, manifest_content=manifest
            )
        )



from os import makedirs
from pathlib import Path


class FileProvider:

    def load(path: str) -> str:
        pass

    @staticmethod
    def mkfile(base_path: str, file_path: str, contents, overwrite=False) -> str:
        file_path = Path(file_path)
        file = Path(base_path).joinpath(file_path)
        file_exists = file.exists()

        if not file_path.parent.exists():
            makedirs(file_path.parent)

        if file_exists and not overwrite:
            return f"File {file_path} exists; doing nothing"

        action = "overwritten" if file_exists and overwrite else "created"

        file.write_text(contents)

        return f"File {file_path} {action}"



#!/usr/bin/env python

import click

from dbt_platform_helper.constants import DEFAULT_TERRAFORM_PLATFORM_MODULES_VERSION
from dbt_platform_helper.domain.config_validator import ConfigValidator
from dbt_platform_helper.domain.pipelines import Pipelines
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.utils.aws import get_codestar_connection_arn
from dbt_platform_helper.utils.click import ClickDocOptGroup
from dbt_platform_helper.utils.git import git_remote
from dbt_platform_helper.utils.versioning import (
    check_platform_helper_version_needs_update,
)


@click.group(chain=True, cls=ClickDocOptGroup)
def pipeline():
    """Pipeline commands."""
    check_platform_helper_version_needs_update()


@pipeline.command()
@click.option(
    "--terraform-platform-modules-version",
    help=f"""Override the default version of terraform-platform-modules with a specific version or branch. 
    Precedence of version used is version supplied via CLI, then the version found in 
    platform-config.yml/default_versions/terraform-platform-modules. 
    In absence of these inputs, defaults to version '{DEFAULT_TERRAFORM_PLATFORM_MODULES_VERSION}'.""",
)
@click.option(
    "--deploy-branch",
    help="""Specify the branch of <application>-deploy used to configure the source stage in the environment-pipeline resource. 
    This is generated from the terraform/environments-pipeline/<aws_account>/main.tf file. 
    (Default <application>-deploy branch is specified in 
    <application>-deploy/platform-config.yml/environment_pipelines/<environment-pipeline>/branch).""",
    default=None,
)
def generate(terraform_platform_modules_version, deploy_branch):
    """
    Given a platform-config.yml file, generate environment and service
    deployment pipelines.

    This command does the following in relation to the environment pipelines:
    - Reads contents of `platform-config.yml/environment-pipelines` configuration.
      The `terraform/environment-pipelines/<aws_account>/main.tf` file is generated using this configuration.
      The `main.tf` file is then used to generate Terraform for creating an environment pipeline resource.

    This command does the following in relation to the codebase pipelines:
    - Generates the copilot pipeline manifest.yml for copilot/pipelines/<codebase_pipeline_name>
    """
    pipelines = Pipelines(
        ConfigProvider(ConfigValidator()),
        git_remote,
        get_codestar_connection_arn,
    )
    pipelines.generate(terraform_platform_modules_version, deploy_branch)
